##
# -*- coding: utf-8 -*-
#
# File:    PsiModProvider.py
# Author:  J. Westbrook
# Date:    19-Mar-2020
# Version: 0.001
#
# Update:
#
#
##
"""
Various utilities for extracting data from the PSI-MOD OBO files
and returning lineage details.
"""

import logging
import os
from collections import OrderedDict

import networkx
import obonet

from rcsb.utils.io.FileUtil import FileUtil

logger = logging.getLogger(__name__)


class PsiModProvider(object):
    """Various utilities for extracting data from the PSI-MOD OBO files
    and returning lineage details.
    """

    def __init__(self, **kwargs):
        urlTarget = kwargs.get("urlTarget", "http://data.bioontology.org/ontologies/PSIMOD/submissions/6/download?apikey=8b5b7825-538d-40e0-9e9e-5ab9274a9aeb")
        urlTargetFallback = "https://github.com/rcsb/py-rcsb_exdb_assets/raw/master/fall_back/psimod.obo"
        dirPath = os.path.join(kwargs.get("cachePath", "."), "psimod")
        useCache = kwargs.get("useCache", True)
        self.__modGraph = self.__reload(urlTarget, urlTargetFallback, dirPath, useCache=useCache)

    def testCache(self):
        if self.__modGraph:
            logger.info("Reading %d nodes and %d edges", len(self.__modGraph), self.__modGraph.number_of_edges())
            if networkx.is_directed_acyclic_graph(self.__modGraph) and len(self.__modGraph) > 1000:
                return True
        return False

    def getVersion(self):
        try:
            return self.__modGraph.graph.get("data-version", "1.013.0")
        except Exception:
            return None

    def exists(self, psimodId):
        try:
            return psimodId in self.__modGraph
        except Exception:
            return False

    def getNode(self, psimodId):
        try:
            return self.__modGraph[psimodId]
        except Exception:
            pass
        return None

    def getRootNodes(self):
        try:
            rootL = [n for n, d in self.__modGraph.out_degree() if d == 0]
            return rootL
        except Exception:
            pass
        return None

    def getName(self, psimodId):
        try:
            return self.__modGraph.nodes[psimodId]["name"]
        except Exception as e:
            logger.debug("Failing %r with %r", psimodId, str(e))
        return None

    def getAdjacentParents(self, psimodId):
        rL = []
        try:
            for child, parent, key in self.__modGraph.out_edges(psimodId, keys=True):
                logger.debug("%r %r - %r -> %r %r", child, self.getName(child), key, parent, self.getName(parent))
                rL.append((child, parent, key))
        except Exception:
            pass
        return rL

    def getPredecessors(self, psimodId):
        rL = []
        try:
            rL = [nd for nd in self.__modGraph.predecessors(psimodId)]
        except Exception:
            pass
        return rL

    def getSuccessors(self, psimodId):
        rL = []
        try:
            rL = [nd for nd in self.__modGraph.successors(psimodId)]
        except Exception:
            pass
        return rL

    def getDescendants(self, psimodId, includeSelf=True):
        linL = []
        try:
            if includeSelf:
                linL.append((psimodId, self.getName(psimodId)))
            for nd in networkx.descendants(self.__modGraph, psimodId):
                logger.debug("%r %r --> %r %r", psimodId, self.getName(psimodId), nd, self.getName(nd))
                linL.append((nd, self.getName(nd)))
        except Exception as e:
            logger.debug("Failing %s with %s", psimodId, str(e))
        return linL

    def getUniqueDescendants(self, psimodIdL, includeSelf=True):
        linL = []
        try:
            ndD = OrderedDict()
            for psimodId in psimodIdL:
                if includeSelf:
                    ndD[psimodId] = True
                for nd in networkx.descendants(self.__modGraph, psimodId):
                    ndD[nd] = True
            #
            for nd in ndD:
                linL.append((nd, self.getName(nd)))
        except Exception as e:
            logger.debug("Failing %s with %s", psimodId, str(e))
        return linL

    def getLineage(self, psiModId):
        linL = []
        lIdL = self.getUniqueDescendants([psiModId])
        for ii, lTup in enumerate(lIdL, 1):
            linL.append((lTup[0], lTup[1], ii))
        return linL

    def exportTreeNodeList(self, psimodIdL):
        """For the input node list export full tree node list including parent nodes

        Args:
            psimodIdL (list): list of covered GO id lists
            includeSelf (bool, optional): include input and parent nodes. Defaults to True.

        Returns:
            list: [{'id': <> 'name': <name> 'parents': [<id>,<id>,...]}]
        """
        trL = []
        try:
            # Generate the full list of nodes and parents -
            ndS = set()
            for psimodId in psimodIdL:
                if not self.exists(psimodId):
                    logger.warning("%s not in current ontology", psimodId)
                    continue
                ndS.add(psimodId)
                for nd in networkx.descendants(self.__modGraph, psimodId):
                    ndS.add(nd)
            #
            for nd in ndS:
                # tupL = [(child, parent, key)]
                pIdL = [tup[1] for tup in self.getAdjacentParents(nd)]
                if not pIdL:
                    logger.info("Node %s (%s) has no parents", nd, self.getName(nd))
                    trL.append({"id": nd, "name": self.getName(nd)})
                else:
                    trL.append({"id": nd, "name": self.getName(nd), "parents": pIdL})

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return trL

    def __reload(self, urlTarget, urlTargetFallback, dirPath, useCache=True):
        """Reload input GO OBO ontology file and return a nx graph object.
        '
                Returns:
                    dictionary[psimodId] = {'name_list': ... , 'id_list': ... 'depth_list': ... }
        """
        goGraph = None
        #
        # mU = MarshalUtil()
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        oboFilePath = os.path.join(dirPath, fn)
        fU.mkdir(dirPath)
        #
        if not useCache:
            for fp in [oboFilePath]:
                try:
                    os.remove(fp)
                except Exception:
                    pass
        #
        if useCache and fU.exists(oboFilePath):
            goGraph = obonet.read_obo(oboFilePath)
        else:
            logger.info("Fetching url %s to resource file %s", urlTarget, oboFilePath)
            ok = fU.get(urlTarget, oboFilePath)
            logger.info("PSIMOD fetch status is %r", ok)
            if not ok:
                ok = fU.get(urlTargetFallback, oboFilePath)
                logger.info("PSIMOD fallback fetch status is %r", ok)
            if ok:
                goGraph = obonet.read_obo(oboFilePath)
        if goGraph:
            logger.info("Reading %d nodes and %d edges", len(goGraph), goGraph.number_of_edges())
        else:
            logger.info("Go graph construction failing")
        #
        return goGraph



##
#  File:           CARDTProvider.py
#  Date:           28-Nov-2020 jdw
#
#  Updated:
#
##
"""
Accessors for CARD molecule data.

"""

import logging
import os.path
import time

import networkx
import obonet

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class CARDProvider:
    """Accessors for CARD molecule data."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(self.__cachePath, "CARD-molecules")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__oD, self.__version = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=350):
        if self.__oD and len(self.__oD) > minCount:
            return True
        else:
            return False

    def __reload(self, dirPath, **kwargs):
        oD = None
        startTime = time.time()
        useCache = kwargs.get("useCache", True)
        #
        cardDumpUrl = kwargs.get("CARDDumpUrl", "https://card.mcmaster.ca/latest/ontology")
        ok = False
        fU = FileUtil()
        cardDumpFileName = "card-ontology.tar.bz2"
        cardDumpPath = os.path.join(dirPath, cardDumpFileName)
        cardDumpDirPath = os.path.join(dirPath, "dump")
        #
        fU.mkdir(dirPath)
        cardDataPath = os.path.join(dirPath, "card-molecules.json")
        #
        logger.info("useCache %r CARDDumpPath %r", useCache, cardDumpPath)
        if useCache and self.__mU.exists(cardDataPath):
            oD = self.__mU.doImport(cardDataPath, fmt="json")
        else:
            logger.info("Fetching url %s path %s", cardDumpUrl, cardDumpPath)
            ok = fU.get(cardDumpUrl, cardDumpPath)
            fU.mkdir(cardDumpDirPath)
            fU.uncompress(cardDumpPath, outputDir=cardDumpDirPath)
            fU.unbundleTarfile(os.path.join(cardDumpDirPath, cardDumpFileName[:-4]), dirPath=cardDumpDirPath)
            logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
            #
            oD = self.__parseCardData(os.path.join(cardDumpDirPath, "aro.obo"))
            ok = self.__mU.doExport(cardDataPath, oD, fmt="json", indent=3)
        # ---
        return oD

    def __parseCardData(self, filePath):
        """Parse CARD ontology file describing antibiotic molecules.

        Args:
            filePath (str): output file path for CARD molecule data

        Returns:
            (dict, version): dictionary of select CARD data, version string


                Note: https://card.mcmaster.ca/aro/0000003
        """
        try:
            oD = {}
            version = None
            cardGraph = obonet.read_obo(filePath)
            logger.info("CARD graph nodes (%d) edges (%d) directed %r", len(cardGraph), cardGraph.number_of_edges(), networkx.is_directed_acyclic_graph(cardGraph))
            for ky, gD in cardGraph.nodes(data=True):
                # logger.info("ky %r gD %r", ky, gD.keys())
                if "xref" in gD:
                    # oD["name"] = gD["name"]
                    # oD["pubChemCId"] = gD["xref"]
                    oD[ky] = {"name": gD["name"], "pubChemCId": gD["xref"], "description": gD["def"]}
            logger.info("Parsed CARD molecule references (%d)", len(oD))
        except Exception as e:
            logger.exception("Failing using %r with %s", filePath, str(e))
        return oD, version



##
# File:    ChemCompModelProvider.py
# Author:  J. Westbrook
# Date:    1-Nov-2018
#
# Updates:
# http://ftp.wwpdb.org/pub/pdb/data/component-models/complete/chem_comp_model.cif.gz
##
"""
Utilities to read resource file containing compilation of CCDC models correspondences
for PDB chemical components.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemCompModelProvider(object):
    """Utilities to read the resource file containing with a compilation of CCDC models correspondences
    for PDB chemical components.
    """

    def __init__(self, **kwargs):
        urlTarget = kwargs.get("urlTarget", "http://ftp.wwpdb.org/pub/pdb/data/component-models/complete/chem_comp_model.cif.gz")
        dirPath = os.path.join(kwargs.get("cachePath", "."), "chem_comp")
        useCache = kwargs.get("useCache", True)
        mappingFileName = kwargs.get("mappingFileName", "ccdc_pdb_mapping.json")
        auditFileName = kwargs.get("mappingFileName", "ccdc_model_audit.json")
        #
        self.__mU = MarshalUtil(workPath=dirPath)
        self.__mappingD, self.__auditD = self.__reload(urlTarget, dirPath, mappingFileName, auditFileName, useCache=useCache)

    def testCache(self):
        logger.info("Lengths map %d", len(self.__mappingD))
        if len(self.__mappingD) > 1000 and len(self.__auditD) > 1000:
            return True
        return False

    def getMapping(self):
        return self.__mappingD

    def getAuditDetails(self):
        return self.__auditD

    def __reload(self, urlTarget, dirPath, mappingFileName, auditFileName, useCache=True):
        """Reload input mmCIF model mapping resource file and return a container list.

        Args:
            urlTarget (str): target url for resource file
            dirPath (str): path to the directory containing cache files
            mappingFileName (str): mapping file name
            auditFileName (str): audit details file name
            useCache (bool, optional): flag to use cached files. Defaults to True.

        Returns:
            (dict,dict): mapping dictionary (pdb_ccId -> CCDC details), audit details (ccId -> audit details)
        """
        mD = {}
        aD = {}
        #
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        filePath = os.path.join(dirPath, fn)
        mappingFilePath = os.path.join(dirPath, mappingFileName)
        auditFilePath = os.path.join(dirPath, auditFileName)
        self.__mU.mkdir(dirPath)
        #
        if not useCache:
            for fp in [filePath, mappingFilePath, auditFilePath]:
                try:
                    os.remove(fp)
                except Exception:
                    pass
        #
        if useCache and fU.exists(mappingFilePath) and fU.exists(auditFilePath):
            mD = self.__mU.doImport(mappingFilePath, fmt="json")
            aD = self.__mU.doImport(auditFilePath, fmt="json")
        else:
            ok = True
            if not (useCache and fU.exists(filePath)):
                logger.info("Fetching url %s for resource file %s", urlTarget, filePath)
                ok = fU.get(urlTarget, filePath)
            if ok:
                cL = self.__mU.doImport(filePath, fmt="mmcif")
                mD = self.__buildMapping(cL)
                ok = self.__mU.doExport(mappingFilePath, mD, fmt="json", indent=3)
                aD = self.__buildAuditDetails(cL)
                ok = self.__mU.doExport(auditFilePath, aD, fmt="json", indent=3)
        #
        return mD, aD

    def __buildMapping(self, cL):
        """Return a dictionary of identfier correspondences for CCDC/CSD models.

        data_M_010_00001
            #
            _pdbx_chem_comp_model.id        M_010_00001
            _pdbx_chem_comp_model.comp_id   010
            #
            _pdbx_chem_comp_model_reference.model_id   M_010_00001
            _pdbx_chem_comp_model_reference.db_name    CSD
            _pdbx_chem_comp_model_reference.db_code    YARXEW
            #
            loop_
            _pdbx_chem_comp_model_feature.model_id
            _pdbx_chem_comp_model_feature.feature_name
            _pdbx_chem_comp_model_feature.feature_value
            M_010_00001 experiment_temperature 123.0
            M_010_00001 publication_doi        10.1021/cg200971f
            M_010_00001 r_factor               3.91
            M_010_00001 all_atoms_have_sites   Y
        #
        """
        rD = {}
        for dataContainer in cL:
            logger.debug("Processing model %r", dataContainer.getName())
            cObj = dataContainer.getObj("pdbx_chem_comp_model")
            ccId = cObj.getValue("comp_id", 0)
            #
            if ccId not in rD:
                rD[ccId] = []

            tObj = dataContainer.getObj("pdbx_chem_comp_model_reference")
            for iRow in range(tObj.getRowCount()):
                dbName = tObj.getValue("db_name", iRow)
                dbCode = tObj.getValue("db_code", iRow)
                rD[ccId].append({"db_name": dbName, "db_code": dbCode})

        return rD

    def __buildAuditDetails(self, cL):
        """Return a dictionary of identfier correspondences for CCDC/CSD models.

                data_M_4YS_00001
                #
                _pdbx_chem_comp_model.id        M_4YS_00001
                _pdbx_chem_comp_model.comp_id   4YS
                #

                _pdbx_chem_comp_model_reference.model_id   M_4YS_00001
                _pdbx_chem_comp_model_reference.db_name    CSD
                _pdbx_chem_comp_model_reference.db_code    CIQKUJ
                #
                loop_
                _pdbx_chem_comp_model_feature.model_id
                _pdbx_chem_comp_model_feature.feature_name
                _pdbx_chem_comp_model_feature.feature_value
                M_4YS_00001 experiment_temperature '290 K'
                M_4YS_00001 publication_doi        10.1107/S1600536807057960
                M_4YS_00001 r_factor               4.230
                M_4YS_00001 csd_version            542
                #
                _pdbx_chem_comp_model_audit.model_id      M_4YS_00001
                _pdbx_chem_comp_model_audit.action_type   'Initial release'
                _pdbx_chem_comp_model_audit.date          2021-02-03

        #
        """
        rD = {}
        for dataContainer in cL:
            logger.debug("Processing model %r", dataContainer.getName())
            cObj = dataContainer.getObj("pdbx_chem_comp_model")
            ccId = cObj.getValue("comp_id", 0)
            modelId = cObj.getValue("id", 0)
            #
            tObj = dataContainer.getObj("pdbx_chem_comp_model_audit")
            aL = []
            for iRow in range(tObj.getRowCount()):
                auditAction = tObj.getValue("action_type", iRow)
                auditDate = tObj.getValue("date", iRow)
                aL.append({"audit_date": auditDate, "action_type": auditAction})

            tObj = dataContainer.getObj("pdbx_chem_comp_model_reference")
            dbName = dbCode = None
            for iRow in range(tObj.getRowCount()):
                dbName = tObj.getValue("db_name", iRow)
                dbCode = tObj.getValue("db_code", iRow)

            rD.setdefault(ccId, []).append({"model_id": modelId, "db_name": dbName, "db_code": dbCode, "audit_list": aL})

        return rD



##
# File:    ResidProvider.py
# Author:  J. Westbrook
# Date:    18-Mar-2020
#
# Updates:
#
##
"""
Utilities to read the RESID resource file and build loadable documents and identifier
    correspondences.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.utils.chemref.ResidReader import ResidReader
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ResidProvider(object):
    """Utilities to read the RESID resource file and build loadable documents and identifier
    correspondences.
    """

    def __init__(self, **kwargs):
        #
        urlTarget = kwargs.get("residUrlTarget", "ftp://ftp.pir.georgetown.edu/pir_databases/other_databases/resid/RESIDUES.XML")
        #
        useCache = kwargs.get("useCache", True)
        #
        dirPath = os.path.join(kwargs.get("cachePath", "."), "resid")
        useCache = kwargs.get("useCache", True)
        residFileName = kwargs.get("residFileName", "resid_correspondences_definitions.json")
        #
        self.__mU = MarshalUtil(workPath=dirPath)
        self.__version, self.__residD = self.__reload(urlTarget, dirPath, residFileName, useCache=useCache)

    def testCache(self):
        logger.info("RESID map length %d", len(self.__residD))
        if len(self.__residD) > 200:
            return True
        return False

    def getComponentIdsOLD(self):
        try:
            return list(self.__residD.keys())
        except Exception:
            pass
        return []

    def getMapping(self):
        return self.__residD

    def getVersion(self):
        return self.__version

    def isChemCompMapped(self, ccId):
        return self.__residD and ccId.upper() in self.__residD

    def __reload(self, urlTarget, dirPath, residFileName, useCache=True):
        """Reload input mmCIF model mapping resource file and return a container list.

        Args:
            urlTarget (str): target url for resource file
            dirPath (str): path to the directory containing cache files
            residFileName (str): mapping file name
            useCache (bool, optional): flag to use cached files. Defaults to True.

        Returns:
            (dict): mapping dictionary (pdb_ccId -> RESID details)
        """
        mD = {}
        version = None
        #
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        filePath = os.path.join(dirPath, fn)
        residFilePath = os.path.join(dirPath, residFileName)
        self.__mU.mkdir(dirPath)
        #
        if not useCache:
            for fp in [filePath, residFilePath]:
                try:
                    os.remove(fp)
                except Exception:
                    pass
        #
        if useCache and fU.exists(residFilePath):
            rD = self.__mU.doImport(residFilePath, fmt="json")
            mD = rD["mapD"]
            version = rD["version"]
        else:
            ok = True
            if not (useCache and fU.exists(filePath)):
                logger.info("Fetching url %s for resource file %s", urlTarget, filePath)
                ok = fU.get(urlTarget, filePath)
            if ok:
                logger.info("Reading %r", filePath)
                xTree = self.__mU.doImport(filePath, fmt="xml")
                rr = ResidReader()
                version, dbObjL = rr.read(xTree)
                mD = self.__buildResidMapping(dbObjL)
                rD = {"mapD": mD, "version": version}
                ok = self.__mU.doExport(residFilePath, rD, fmt="json", indent=3)

        #
        return version, mD

    def __getPdbCrossRef(self, xRefList):
        ccId = None
        for xRef in xRefList:
            if xRef.startswith("PDBHET"):
                ccId = xRef.split(":")[1]
                break
        return ccId

    def __buildResidMapping(self, dbObjL):
        """Return a dictionary of RESID to chemical component mapping.

        doc = {
            "residCode": entryElement.findtext("{ns}Header/{ns}Code".format(ns=self.__ns)),
            "names": [name.text for name in entryElement.findall("{ns}Names/{ns}Name".format(ns=self.__ns))],
            "nameXrefs": [xref.text for xref in entryElement.findall("{ns}Names/{ns}Xref".format(ns=self.__ns))],
            "ontRefs": [xref.text for xref in entryElement.findall("{ns}SequenceCode/{ns}Xref".format(ns=self.__ns))],
            "genEnzymes": [xref.text for xref in entryElement.findall("{ns}GeneratingEnzyme/{ns}EnzymeName".format(ns=self.__ns))],
            "features": [feature.text for feature in entryElement.findall("{ns}Features/{ns}Feature".format(ns=self.__ns))],
        }
        """
        retD = {}
        logger.info("Processing (%d) RESID entries", len(dbObjL))
        for dbObj in dbObjL:
            ccId = self.__getPdbCrossRef(dbObj["nameXrefs"])
            if not ccId:
                continue
            #
            residCode = dbObj["residCode"]
            residName = dbObj["names"][0]
            related = []
            for xRef in dbObj["nameXrefs"]:
                db = xRef.split(":")[0]
                if db == "PDBHET":
                    continue
                dbCode = xRef.split(":")[1]
                #
                related.append({"resourceName": db, "resourceCode": dbCode})
            #
            modResL = []
            for feature in dbObj["features"]:
                if feature.startswith("MOD_RES"):
                    modResL.append(feature[len("MOD_RES") :].strip())
            genEnzymes = sorted(set(dbObj["genEnzymes"]))
            #
            ontRefs = [t[4:] for t in sorted(set(dbObj["ontRefs"])) if t.startswith("PSI-")]
            retD.setdefault(ccId, []).append(
                {"residCode": residCode, "residName": residName, "relatedResource": related, "modRes": modResL, "genEnzymes": genEnzymes, "ontRefs": ontRefs}
            )

        return retD



##
# File: DrugBankProvider.py
# Author:  J. Westbrook
# Date:  8-Aug-2019
#
# Build loadable correspondence data from DrugBank.
#
# Update:
#
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import os
import time

from rcsb.utils.chemref.DrugBankReader import DrugBankReader
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class DrugBankProvider(object):
    """Utilities to read the DrugBank resource file and build loadable documents and identifier
    correspondences.
    """

    def __init__(self, **kwargs):
        #
        self.__dbMapD, self.__dbObjL = self.__reload(**kwargs)
        self.__version = None
        #

    def testCache(self):
        try:
            logger.info("Lengths map %d objl %d", len(self.__dbMapD["id_map"]), len(self.__dbObjL))
            if (len(self.__dbMapD["id_map"]) > 340) and (len(self.__dbObjL) > 995):
                return True
        except Exception as e:
            logger.debug("Failing with %s", str(e))

        return False

    def getMapping(self):
        return self.__dbMapD

    def getVersion(self):
        return self.__dbMapD["version"] if self.__dbMapD and "version" in self.__dbMapD else None

    def getDocuments(self, mapD=None):
        """[summary]

        Args:
            mapD ([type], optional): mapping dictionary {drugbank_id: ccId}. Defaults to None.

        Returns:
            (list): loadable DrugBank documents
        """
        if mapD:
            return self.__buildDocuments(self.__dbObjL, mapD)
        else:
            dbIdD = {}
            if "id_map" in self.__dbMapD:
                for ccId, dD in self.__dbMapD["id_map"].items():
                    if "drugbank_id" in dD:
                        dbIdD[dD["drugbank_id"]] = ccId
            return self.__buildDocuments(self.__dbObjL, dbIdD)

    def __reload(self, **kwargs):
        """Reload DrugBank mapping data and optionally supporting repository data file.

        Args:
            urlTarget (str): target url for resource file
            dirPath (str): path to the directory containing cache files
            mappingFileName (str): mapping file name
            docListFileName (str): document list file name
            useCache (bool, optional): flag to use cached files. Defaults to True.
            useDownload (bool, optional): flag to use a downloaded DrugBank resource file. Default to True.

        Returns:
            (dict, list): identifiers mapping dictionary (pdb_ccId -> DrugBank details), DrugBank object list
        """
        startTime = time.time()
        logger.info("Starting db reload at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        # Latest url seems to broken on ~ Sep 19, 2020
        urlTarget = kwargs.get("urlTarget", "https://go.drugbank.com/releases/latest/downloads/all-full-database")
        # urlTarget = kwargs.get("urlTarget", "https://go.drugbank.com/releases/5-1-7/downloads/all-full-database")
        dirPath = os.path.join(kwargs.get("cachePath", "."), "DrugBank")
        useCache = kwargs.get("useCache", True)
        useDownload = kwargs.get("useDownload", True)
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)
        mappingFileName = kwargs.get("mappingFileName", "drugbank_pdb_mapping.json")
        docListFileName = kwargs.get("docListFileName", "drugbank_documents.pic")
        mU = MarshalUtil(workPath=dirPath)
        #
        dbMapD = {}
        dbObjL = []
        fU = FileUtil()
        mappingFilePath = os.path.join(dirPath, mappingFileName)
        docListFilePath = os.path.join(dirPath, docListFileName)
        filePath = os.path.join(dirPath, "full database.xml")
        mU.mkdir(dirPath)
        #
        if not useCache:
            for fp in [filePath, mappingFilePath, docListFilePath]:
                try:
                    os.remove(fp)
                except Exception:
                    pass
        #
        if useCache and fU.exists(mappingFilePath) and fU.exists(docListFilePath):
            logger.debug("Using cached %r", mappingFilePath)
            dbMapD = mU.doImport(mappingFilePath, fmt="json")
            dbObjL = mU.doImport(docListFilePath, fmt="pickle")
            # done all cached -
            endTime = time.time()
            logger.info(
                "Completed cache recovery (%d/%d) at %s (%.4f seconds)",
                len(dbObjL),
                len(dbMapD["id_map"]),
                time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                endTime - startTime,
            )
            return dbMapD, dbObjL
        #
        ok = fU.exists(filePath)
        if not ok:
            if not username or not password:
                logger.warning("Missing credentials for DrugBank file download...")
            zipFilePath = os.path.join(dirPath, "full_database.zip")
            if useDownload and fU.exists(zipFilePath):
                logger.info("Using existing downloaded file %r", zipFilePath)
            else:
                logger.info("Fetching url %s to resource file %s", urlTarget, filePath)
                ok = fU.get(urlTarget, zipFilePath, username=username, password=password)
                endTime = time.time()
                logger.info("Completed db fetch at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
            #
            fp = fU.uncompress(zipFilePath, outputDir=dirPath)
            ok = fp.endswith("full database.xml")
            endTime = time.time()
            logger.info("Completed unzip at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)

        if ok:
            logger.debug("Reading %r", filePath)
            xTree = mU.doImport(filePath, fmt="xml")
            endTime = time.time()
            logger.info("Completed xml read at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
            dbr = DrugBankReader()
            version, dbObjL = dbr.read(xTree)
            endTime = time.time()
            logger.info("Completed parsing (%d) (%r) at %s (%.4f seconds)", len(dbObjL), version, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)

            dbMapD = self.__buildMapping(dbObjL)
            dbMapD["version"] = version
            ok = mU.doExport(mappingFilePath, dbMapD, fmt="json", indent=3, enforceAscii=False)
            ok = mU.doExport(docListFilePath, dbObjL, fmt="pickle")
            endTime = time.time()
            logger.info(
                "Completed db %d/%d processing at %s (%.4f seconds)", len(dbObjL), len(dbMapD["id_map"]), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime
            )
        else:
            logger.error("Drugbank resource file missing %r", fp)
        #
        return dbMapD, dbObjL

    def __buildDocuments(self, dbObjL, dbIdD=None):
        """Build loadable documents subject to corresponding identifiers in the input mapping dictionary.

        Args:
            dbObjL (list): list of extracted DrugBank objects
            dbIdD (dict, optional): dictionary of DrugBank identifier mapping. Defaults to None.

        Returns:
            (list): DrugBank loadable documents
        """
        dbIdD = dbIdD if dbIdD else {}
        dbDocL = []
        for dbObj in dbObjL:
            if "drugbank_id" in dbObj:
                if dbIdD and dbObj["drugbank_id"] not in dbIdD:
                    continue
                lD = self.__buildDocument(dbObj)
                dbDocL.append(lD)
            else:
                logger.debug("dbObj.keys() %r", list(dbObj.keys()))
        return dbDocL
        #

    def __buildDocument(self, dbObj):
        """Construct local loadable object from input DrugBank extracted data object
        conforming to the following category organization:

         _drugbank_info.drugbank_id
         _drugbank_info.name
         _drugbank_info.description
         _drugbank_info.synonyms
         _drugbank_info.brand_names
         _drugbank_info.affected_organisms
         _drugbank_info.indication
         _drugbank_info.pharmacology
         _drugbank_info.mechanism_of_action
         _drugbank_info.cas_number
         _drugbank_info.drug_categories
         _drugbank_info.drug_groups
         _drugbank_info.atc_codes


         _drugbank_target.ordinal
         _drugbank_target.name
         _drugbank_target.interaction_type
         _drugbank_target.target_actions
         _drugbank_target.organism_common_name
         _drugbank_target.reference_database_name
         _drugbank_target.reference_database_accession_code
         _drugbank_target.seq_one_letter_code

        """
        oD = {}
        # oD["_drugbank_id"] = dbObj["drugbank_id"]
        oD["drugbank_container_identifiers"] = {"drugbank_id": dbObj["drugbank_id"]}

        dbiD = {}
        textKeys = [
            ("drugbank_id", "drugbank_id"),
            ("name", "name"),
            ("description", "description"),
            ("indication", "indication"),
            ("pharmacology", "pharmacology"),
            ("mechanism_of_action", "mechanism_of_action"),
            ("cas_number", "cas_number"),
        ]
        listKeys = [
            ("drug_categories", "drug_categories"),
            ("groups", "drug_groups"),
            ("aliases", "synonyms"),
            ("products", "brand_names"),
            ("affected_organisms", "affected_organisms"),
            ("atc_codes", "atc_codes"),
        ]
        # For category drugbank_info
        for textKey, docKey in textKeys:
            if textKey in dbObj and dbObj[textKey]:
                dbiD[docKey] = dbObj[textKey].replace("\r", "").replace("\n", " ")
        #
        for listKey, docKey in listKeys:
            if listKey in dbObj and dbObj[listKey]:
                dbiD[docKey] = list(dbObj[listKey])
            else:
                logger.debug("MISSING KEY %s", listKey)
        #
        oD["drugbank_info"] = dbiD
        #
        # For category drugbank_target -
        #
        tL = []
        if "target_interactions" in dbObj:
            for ii, tid in enumerate(dbObj["target_interactions"], 1):
                tD = {}
                tD["ordinal"] = ii
                tD["interaction_type"] = tid["category"]
                tD["name"] = tid["name"]
                if "organism" in tid:
                    tD["organism_common_name"] = tid["organism"]
                if "actions" in tid and tid["actions"]:
                    tD["target_actions"] = tid["actions"]
                if "amino-acid-sequence" in tid and tid["amino-acid-sequence"] and tid["amino-acid-sequence"]:
                    tD["seq_one_letter_code"] = "".join(tid["amino-acid-sequence"].split("\n")[1:])
                if "uniprot_ids" in tid and tid["uniprot_ids"]:
                    tD["reference_database_name"] = "UniProt"
                    tD["reference_database_accession_code"] = tid["uniprot_ids"]
                #
                tL.append(tD)
        #
        if tL:
            oD["drugbank_target"] = tL
        #
        return oD

    def __buildMapping(self, dbObjL):
        """Build PDB->DrugBank identifier correspondences.

        Args:
            dbObjL (list): list of extracted DrugBank data objects

        Returns:
            (dict): dictionary DrugBank identifier correspondences
        """
        logger.info("DrugBank full record length %d", len(dbObjL))
        dbMapD = {}
        mD = {}
        for dD in dbObjL:
            dbId = dD["drugbank_id"]
            pdbIds = ""
            if "external_identifiers" in dD:
                for exD in dD["external_identifiers"]:
                    if exD["resource"] == "PDB":
                        logger.debug("dbId %s pdbids %r ccids %r", dbId, pdbIds, exD["identifier"])
                        if exD["identifier"] not in mD:
                            mD[exD["identifier"]] = []
                        mD[exD["identifier"]] = {"drugbank_id": dbId, "aliases": list(dD["aliases"])}
                        #
                        if "atc_codes" in dD and dD["atc_codes"]:
                            mD[exD["identifier"]]["atc_codes"] = dD["atc_codes"]
                        #
                        if "products" in dD and dD["products"]:
                            mD[exD["identifier"]]["brand_names"] = dD["products"]
                        #
                        if "target_interactions" in dD:
                            for tid in dD["target_interactions"]:
                                tD = {}
                                tD["type"] = tid["category"]
                                tD["name"] = tid["name"]
                                tD["organism"] = tid["organism"]
                                if tid["actions"]:
                                    tD["actions"] = tid["actions"]
                                if tid["known_action"]:
                                    tD["known_action"] = tid["known_action"]
                                if "uniprot_ids" in tid:
                                    tD["uniprot_ids"] = tid["uniprot_ids"]
                                #
                                if "target_interactions" not in mD[exD["identifier"]]:
                                    mD[exD["identifier"]]["target_interactions"] = []
                                mD[exD["identifier"]]["target_interactions"].append(tD)
        logger.info("Match length is %d", len(mD))
        dbMapD["id_map"] = mD
        #
        inD = {}
        atcD = {}
        for dD in dbObjL:
            dbId = dD["drugbank_id"]
            if "inchikey" in dD and dD["inchikey"] and len(dD["inchikey"]) > 13:
                if dD["inchikey"] not in inD:
                    inD[dD["inchikey"]] = []
                inD[dD["inchikey"]].append({"drugbank_id": dbId, "inchikey": dD["inchikey"], "name": dD["name"]})
            #
            if "atc_codes" in dD and dD["atc_codes"]:
                atcD[dbId] = dD["atc_codes"]
        logger.info("Drugbank InChIKey dictionary length %d", len(inD))
        logger.info("Drugbank ATC  dictionary length %d", len(atcD))
        #
        dbMapD["inchikey_map"] = inD
        dbMapD["db_atc_map"] = atcD
        return dbMapD



##
#  File:           DrugCentralProvider.py
#  Date:           11-Nov-2020 jdw
#
#  Updated:
#
##
"""
Accessors for DrugCentral small molecule data.

"""

import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class DrugCentralProvider:
    """Accessors for DrugCentral small molecule data."""

    def __init__(self, **kwargs):
        #
        self.__version = "0.50"
        cachePath = kwargs.get("cachePath", ".")
        useCache = kwargs.get("useCache", True)
        self.__dirPath = os.path.join(cachePath, "CACHE", "DrugCentral")
        drugCentralUrl = "http://unmtid-shinyapps.net/download/structures.smiles.tsv"
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__retD = self.__reload(drugCentralUrl, self.__dirPath, useCache=useCache)
        #

    def testCache(self):
        logger.info("Length %d", len(self.__retD))
        return self.__retD and len(self.__retD) > 3900

    def getInChIKey(self, drugCentralId):
        rVal = None
        try:
            rVal = self.__retD[drugCentralId]["inchikey"]
        except Exception:
            pass
        return rVal

    def getSMILES(self, drugCentralId):
        rVal = None
        try:
            rVal = self.__retD[drugCentralId]["smiles"]
        except Exception:
            pass
        return rVal

    def getCAS(self, drugCentralId):
        rVal = None
        try:
            rVal = self.__retD[drugCentralId]["cas"]
        except Exception:
            pass
        return rVal

    def getINN(self, drugCentralId):
        rVal = None
        try:
            rVal = self.__retD[drugCentralId]["inn"]
        except Exception:
            pass
        return rVal

    def __reload(self, drugCentralUrl, dirPath, useCache=True, fmt="json"):
        startTime = time.time()
        ok = False
        fU = FileUtil()
        fU.mkdir(dirPath)
        #
        drugCentralDataPath = os.path.join(dirPath, "drugcentral_data." + fmt)
        if useCache and self.__mU.exists(drugCentralDataPath):
            logger.info("useCache %r using %r", useCache, drugCentralDataPath)
            retD = self.__mU.doImport(drugCentralDataPath, fmt=fmt)
            ok = True
        else:
            #
            inpFileName = fU.getFileName(drugCentralUrl)
            inpPath = os.path.join(dirPath, inpFileName)
            logger.info("Fetching url %s path %s", drugCentralUrl, inpPath)
            ok = fU.get(drugCentralUrl, inpPath)
            logger.info("Completed fetches at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
            if ok:
                retD = self.__drugCentralParse(inpPath)
                self.__mU.doExport(drugCentralDataPath, retD, fmt=fmt, indent=3)
        return retD

    def __drugCentralParse(self, filePath):
        #     Fields:   SMILES  InChI   InChIKey        ID      INN     CAS_RN
        retD = {}
        rowL = self.__mU.doImport(filePath, fmt="tdd", rowFormat="list")
        logger.info("DrugCentral data length %d", len(rowL))
        #
        for row in rowL[1:]:
            lenR = len(row)
            if lenR < 3:
                logger.debug("Bad row %r", row)
                continue
            casRn = row[5] if lenR > 5 else None
            inn = row[4] if lenR > 4 else None
            retD[row[3]] = {"smiles": row[0], "inchiKey": row[2], "cas": casRn, "inn": inn}
        return retD



##
# File:    ChemCompProvider.py
# Author:  J. Westbrook
# Date:    22-Nov-2019
#
# Updates:
#
##
"""
Utilities to provide essential data items for chemical component definitions.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemCompProvider(object):
    """Utilities to provide essential data items for chemical component definitions.
    """

    def __init__(self, **kwargs):
        urlTarget = kwargs.get("ccUrlTarget", "http://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif.gz")
        # self.__birdUrlTarget = kwargs.get("birdUrlTarget", "ftp://ftp.wwpdb.org/pub/pdb/data/bird/prd/prdcc-all.cif.gz")
        #
        dirPath = os.path.join(kwargs.get("cachePath", "."), "chem_comp")
        useCache = kwargs.get("useCache", True)
        ccdFileName = kwargs.get("ccdFileName", "ccd_abbridged_definitions.json")
        #
        self.__mU = MarshalUtil(workPath=dirPath)
        self.__ccdD = self.__reload(urlTarget, dirPath, ccdFileName, useCache=useCache)

    def testCache(self):
        logger.info("Lengths map %d", len(self.__ccdD))
        if len(self.__ccdD) > 30000:
            return True
        return False

    def getParentComponent(self, ccId):
        try:
            tS = self.__ccdD[ccId]["mon_nstd_parent_comp_id"]
            return tS if tS and tS not in [".", "?"] else None
        except Exception:
            pass
        return None

    def getType(self, ccId):
        try:
            tS = self.__ccdD[ccId]["type"]
            return tS if tS and tS not in [".", "?"] else None
        except Exception:
            pass
        return None

    def getSubComponentList(self, ccId):
        try:
            tS = self.__ccdD[ccId]["pdbx_subcomponent_list"]
            return tS if tS and tS not in [".", "?"] else None
        except Exception:
            pass
        return None

    def getComponentIds(self):
        try:
            return list(self.__ccdD.keys())
        except Exception:
            pass
        return []

    def getAbbridged(self):
        return self.__ccdD

    def __reload(self, urlTarget, dirPath, ccdFileName, useCache=True):
        """Reload input mmCIF model mapping resource file and return a container list.

        Args:
            urlTarget (str): target url for resource file
            dirPath (str): path to the directory containing cache files
            ccdFileName (str): mapping file name
            useCache (bool, optional): flag to use cached files. Defaults to True.

        Returns:
            (dict): mapping dictionary (pdb_ccId -> CCDC details)
        """
        mD = {}
        #
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        filePath = os.path.join(dirPath, fn)
        ccdFilePath = os.path.join(dirPath, ccdFileName)
        self.__mU.mkdir(dirPath)
        #
        if not useCache:
            for fp in [filePath, ccdFilePath]:
                try:
                    os.remove(fp)
                except Exception:
                    pass
        #
        if useCache and fU.exists(ccdFilePath):
            mD = self.__mU.doImport(ccdFilePath, fmt="json")
        else:
            ok = True
            if not (useCache and fU.exists(filePath)):
                logger.info("Fetching url %s for resource file %s", urlTarget, filePath)
                ok = fU.get(urlTarget, filePath)
            if ok:
                cL = self.__mU.doImport(filePath, fmt="mmcif")
                mD = self.__buildAbbridged(cL)
                ok = self.__mU.doExport(ccdFilePath, mD, fmt="json", indent=3)
        #
        return mD

    def __buildAbbridged(self, cL):
        """Return a dictionary of abbridged CCD info.
        """
        atNameList = [
            "id",
            "name",
            "type",
            "pdbx_type",
            "formula",
            "mon_nstd_parent_comp_id",
            "pdbx_synonyms",
            "pdbx_formal_charge",
            "pdbx_initial_date",
            "pdbx_modified_date",
            "pdbx_ambiguous_flag",
            "pdbx_release_status",
            "pdbx_replaced_by",
            "pdbx_replaces",
            "formula_weight",
            "one_letter_code",
            "three_letter_code",
        ]
        retD = {}
        for dataContainer in cL:
            tD = {}
            logger.debug("Processing model %r", dataContainer.getName())
            cObj = dataContainer.getObj("chem_comp")
            ccId = cObj.getValue("id", 0)
            for atName in atNameList:
                tD[atName] = cObj.getValueOrDefault(atName, 0, defaultValue=None)
            retD[ccId] = tD
        return retD



##
#  File:           ChEMBLProvider.py
#  Date:           9-Nov-2020 jdw
#
#  Updated:
#  2-Dec-2020 jdw Add ChEMBL API access methods
##
"""
Accessors for ChEMBL small molecule data.

"""
# pylint: disable=line-too-long
import logging
import os.path
import time

from chembl_webresource_client.new_client import new_client
from chembl_webresource_client.unichem import unichem_client

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChEMBLProvider:
    """Accessors for ChEMBL small molecule data."""

    def __init__(self, **kwargs):
        #
        self.__version = "0.50"
        cachePath = kwargs.get("cachePath", ".")
        useCache = kwargs.get("useCache", True)
        self.__dirPath = os.path.join(cachePath, "CACHE", "ChEMBL")

        chemblDbUrl = kwargs.get("ChEMBLDbUrl", "ftp://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/latest/")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__retD = self.__reload(chemblDbUrl, self.__dirPath, useCache=useCache)
        #

    def testCache(self):
        return self.__retD and len(self.__retD) > 10000

    def getInChIKey(self, chemblId):
        rVal = None
        try:
            rVal = self.__retD[chemblId]["inchikey"]
        except Exception:
            pass
        return rVal

    def getSMILES(self, chemblId):
        rVal = None
        try:
            rVal = self.__retD[chemblId]["smiles"]
        except Exception:
            pass
        return rVal

    def __reload(self, chemblDbUrl, dirPath, useCache=True, fmt="json"):
        startTime = time.time()
        ok = False
        fU = FileUtil()
        fU.mkdir(dirPath)
        #
        baseVersion = 27
        # ChEMBL current version 27,...
        # template:  chembl_27_chemreps.txt.gz
        #
        inpFileName = "chembl_" + str(baseVersion) + ".fa.gz"
        #

        chemblDataPath = os.path.join(dirPath, "chembl_data." + fmt)

        #
        if useCache and self.__mU.exists(chemblDataPath):
            logger.info("useCache %r using %r", useCache, chemblDataPath)
            retD = self.__mU.doImport(chemblDataPath, fmt=fmt)
            ok = True
        else:
            #
            for vers in range(baseVersion, baseVersion + 10):
                inpFileName = "chembl_" + str(vers) + "_chemreps.txt.gz"
                inpPath = os.path.join(dirPath, inpFileName)
                url = os.path.join(chemblDbUrl, inpFileName)
                ok = fU.get(url, inpPath)
                logger.info("Fetching url %s path %s", url, inpPath)
                if ok:
                    break
            #
            logger.info("Completed fetches at %s (%.4f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
            retD = self.__chemblParse(inpPath)
            self.__mU.doExport(chemblDataPath, retD, fmt=fmt, indent=3)
        return retD

    def __chemblParse(self, filePath):
        retD = {}
        rowL = self.__mU.doImport(filePath, fmt="tdd", rowFormat="list")
        logger.info("ChEMBL data length %d", len(rowL))
        for row in rowL[1:]:
            if len(row) < 4:
                logger.debug("Bad row %r", row)
                continue
            retD[row[0]] = {"smiles": row[1], "inchiKey": row[3]}
        return retD

    def getDrugData(self, moleculeChEMBLIdList):
        """Get drug data for the input ChEMBL molecule list.

        Args:
            moleculeChEMBLIdList (list): list of ChEMBL molecule identifiers

        Returns:
          (dict):  dictionary  {ChEMBId: {drug data}}

        """
        oD = {}
        chunkSize = 50
        try:
            for ii in range(0, len(moleculeChEMBLIdList), chunkSize):
                drug = new_client.drug  # pylint: disable=no-member
                drug.set_format("json")
                mDL = drug.filter(molecule_chembl_id__in=moleculeChEMBLIdList[ii : ii + chunkSize])
                if mDL:
                    logger.info("mDL (%d)", len(mDL))
                    for mD in mDL:
                        oD.setdefault(mD["molecule_chembl_id"], []).append(mD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oD

    def getMoleculeData(self, moleculeChEMBLIdList):
        """Get molecule data for the input ChEMBL molecule list.

        Args:
            moleculeChEMBLIdList (list): list of ChEMBL molecule identifiers

        Returns:
          (dict):  dictionary  {ChEMBId: {molecule data}}

        """
        oD = {}
        chunkSize = 50
        try:
            for ii in range(0, len(moleculeChEMBLIdList), chunkSize):
                drug = new_client.molecule  # pylint: disable=no-member
                drug.set_format("json")
                mDL = drug.filter(molecule_chembl_id__in=moleculeChEMBLIdList[ii : ii + chunkSize])
                if mDL:
                    logger.info("mDL (%d)", len(mDL))
                    for mD in mDL:
                        oD.setdefault(mD["molecule_chembl_id"], []).append(mD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oD

    def getMoleculeByInChIKey(self, inchiKeyList):
        """Get molecule data for the input InChI Key list.

        Args:
            inchiKeyList (list): list of InChI Key identifiers

        Returns:
          (dict):  dictionary  {ChEMBId: {molecule data}}

        """
        oD = {}
        chunkSize = 50
        try:
            for ii in range(0, len(inchiKeyList), chunkSize):
                drug = new_client.molecule  # pylint: disable=no-member
                drug.set_format("json")
                mDL = drug.get(inchiKeyList[ii : ii + chunkSize])
                if mDL:
                    logger.info("mDL (%d)", len(mDL))
                    for mD in mDL:
                        oD.setdefault(mD["molecule_chembl_id"], []).append(mD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oD

    def getUniChemData(self, inchiKeyList):
        """Get UniChem data for the input InChiKey list.

        Args:
            InChIList (list): list of InChI key molecule identifiers

        Returns:
          (dict):  dictionary  {ChEMBId: {molecule data}}

        """
        mapD = {
            1: {"name": "chembl", "baseUrl": "https://www.ebi.ac.uk/chembl/", "entryUrl": "https://www.ebi.ac.uk/chembldb/compound/inspect/"},
            3: {"name": "pdb", "baseUrl": "http://www.ebi.ac.uk/pdbe/", "entryUrl": "http://www.ebi.ac.uk/pdbe-srv/pdbechem/chemicalCompound/show/"},
            2: {"name": "drugbank", "baseUrl": "http://drugbank.ca/", "entryUrl": "http://www.drugbank.ca/drugs/"},
            5: {"name": "pubchem_dotf", "baseUrl": "http://pubchem.ncbi.nlm.nih.gov/sources/sources.cgi", "entryUrl": "http://pubchem.ncbi.nlm.nih.gov/substance/"},
            4: {"name": "gtopdb", "baseUrl": "http://www.guidetopharmacology.org", "entryUrl": "http://www.guidetopharmacology.org/GRAC/LigandDisplayForward?ligandId="},
            11: {"name": "ibm", "baseUrl": "http://www-935.ibm.com/services/us/gbs/bao/siip/nih/", "entryUrl": "http://www-935.ibm.com/services/us/gbs/bao/siip/nih/?sid="},
            6: {"name": "kegg_ligand", "baseUrl": "http://www.genome.jp/kegg/ligand.html", "entryUrl": "http://www.genome.jp/dbget-bin/www_bget?"},
            9: {"name": "zinc", "baseUrl": "http://zinc15.docking.org", "entryUrl": "http://zinc15.docking.org/substances/"},
            8: {"name": "nih_ncc", "baseUrl": "http://nihsmr.evotec.com/evotec/", "entryUrl": ""},
            10: {"name": "emolecules", "baseUrl": "https://www.emolecules.com/", "entryUrl": "https://www.emolecules.com/cgi-bin/more?vid="},
            12: {"name": "atlas", "baseUrl": "http://www.ebi.ac.uk/gxa/home", "entryUrl": "http://www.ebi.ac.uk/gxa/query?conditionQuery="},
            7: {"name": "chebi", "baseUrl": "http://www.ebi.ac.uk/chebi/downloadsForward.do", "entryUrl": "http://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI%3A"},
            14: {
                "name": "fdasrs",
                "baseUrl": "http://fdasis.nlm.nih.gov/srs/srs.jsp",
                "entryUrl": "http://fdasis.nlm.nih.gov/srs/ProxyServlet?mergeData=true&objectHandle=DBMaint&APPLICATION_NAME=fdasrs&actionHandle=default&nextPage=jsp/srs/ResultScreen.jsp&TXTSUPERLISTID=",
            },
            15: {"name": "surechembl", "baseUrl": "https://www.surechembl.org/search/", "entryUrl": "https://www.surechembl.org/chemical/"},
            21: {"name": "pubchem_tpharma", "baseUrl": "http://www.thomson-pharma.com/", "entryUrl": "http://pubchem.ncbi.nlm.nih.gov/substance/"},
            22: {"name": "pubchem", "baseUrl": "http://pubchem.ncbi.nlm.nih.gov", "entryUrl": "http://pubchem.ncbi.nlm.nih.gov/compound/"},
            27: {"name": "recon", "baseUrl": "https://vmh.uni.lu", "entryUrl": "https://vmh.uni.lu/"},
            28: {"name": "molport", "baseUrl": "https://www.molport.com/shop/index", "entryUrl": "https://www.molport.com/shop/molecule-link/"},
            31: {
                "name": "bindingdb",
                "baseUrl": "https://www.bindingdb.org/bind/index.jsp",
                "entryUrl": "http://www.bindingdb.org/bind/chemsearch/marvin/MolStructure.jsp?monomerid=",
            },
            41: {"name": "swisslipids", "baseUrl": "http://www.swisslipids.org/", "entryUrl": "http://www.swisslipids.org/"},
            29: {"name": "nikkaji", "baseUrl": "http://jglobal.jst.go.jp/en/", "entryUrl": "http://jglobal.jst.go.jp/en/redirect?Nikkaji_No="},
            32: {"name": "comptox", "baseUrl": "https://comptox.epa.gov/dashboard/", "entryUrl": "https://comptox.epa.gov/dashboard/"},
            33: {"name": "lipidmaps", "baseUrl": "http://www.lipidmaps.org", "entryUrl": "http://www.lipidmaps.org/data/LMSDRecord.php?LMID="},
            35: {"name": "carotenoiddb", "baseUrl": "http://carotenoiddb.jp/index.html", "entryUrl": "http://carotenoiddb.jp/Entries/"},
            36: {"name": "metabolights", "baseUrl": "http://www.ebi.ac.uk/metabolights/", "entryUrl": "http://www.ebi.ac.uk/metabolights/"},
            37: {"name": "brenda", "baseUrl": "https://www.brenda-enzymes.org/index.php", "entryUrl": "https://www.brenda-enzymes.org/ligand.php?brenda_ligand_id="},
            17: {"name": "pharmgkb", "baseUrl": "https://www.pharmgkb.org", "entryUrl": "https://www.pharmgkb.org/drug/"},
            18: {"name": "hmdb", "baseUrl": "http://www.hmdb.ca", "entryUrl": "http://www.hmdb.ca/metabolites/"},
            24: {
                "name": "nmrshiftdb2",
                "baseUrl": "http://nmrshiftdb.nmr.uni-koeln.de/portal/media-type/html/user/anon/page/default.psml/js_pane/P-Home",
                "entryUrl": "http://nmrshiftdb.org/molecule/",
            },
            25: {"name": "lincs", "baseUrl": "http://www.lincsproject.org/", "entryUrl": "http://identifiers.org/lincs.smallmolecule/"},
            39: {"name": "chemicalbook", "baseUrl": "https://www.chemicalbook.com", "entryUrl": "https://www.chemicalbook.com/ChemicalProductProperty_EN_"},
            20: {"name": "selleck", "baseUrl": "http://www.selleckchem.com", "entryUrl": "http://www.selleckchem.com/products/"},
            23: {"name": "mcule", "baseUrl": "https://mcule.com", "entryUrl": "https://mcule.com/"},
            26: {"name": "actor", "baseUrl": "https://actor.epa.gov", "entryUrl": "http://actor.epa.gov/actor/chemical.xhtml?casrn="},
            34: {"name": "drugcentral", "baseUrl": "http://drugcentral.org", "entryUrl": "http://drugcentral.org/drugcard/"},
            38: {"name": "rhea", "baseUrl": "http://www.rhea-db.org", "entryUrl": "http://www.rhea-db.org/searchresults?q=CHEBI:"},
        }
        oD = {}
        try:
            for ky in inchiKeyList:
                unc = unichem_client  # pylint: disable=no-member
                # unc.set_format("json")
                uDL = unc.get(ky)
                if uDL:
                    qD = {}
                    for uD in uDL:
                        if "src_id" in uD and int(uD["src_id"]) in mapD:
                            qD[mapD[int(uD["src_id"])]["name"]] = uD["src_compound_id"]
                    if qD:
                        oD[ky] = qD

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oD



##
# File:    PubChemUtils.py
# Date:    1-May-2020
#
# Updates:
#
#   28-May-2020 jdw use alternative url fetch library.
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

# pylint: disable=too-many-lines

import logging
import os
import time
from collections import namedtuple, defaultdict

from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.UrlRequestUtil import UrlRequestUtil

try:
    from urllib.parse import quote
except ImportError:
    from urllib2 import quote

logger = logging.getLogger(__name__)

ExternalAnnotationFields = ("featureType", "name", "value", "description", "dataType", "units", "provReferenceId", "provSourceName")
ExternalAnnotation = namedtuple("ExternalAnnotation", ExternalAnnotationFields, defaults=(None,) * len(ExternalAnnotationFields))

ChemicalIdentifierFields = ("idCode", "identifierSource", "identifierType", "identifier", "indexName")
ChemicalIdentifier = namedtuple("ChemicalIdentifier", ChemicalIdentifierFields, defaults=(None,) * len(ChemicalIdentifierFields))


class PubChemUtils(object):
    """Manage search and fetch queries for PubChem compound, substance and assay data and related annotations."""

    def __init__(self, **kwargs):
        self.__verbose = kwargs.get("verbose", False)
        self.__delaySeconds = float(kwargs.get("delaySeconds", "0.15"))
        self.__urlPrimary = kwargs.get("urlPrimary", "https://pubchem.ncbi.nlm.nih.gov")
        #

    def assemble(self, chemicalIdentifier, **kwargs):
        """Build a PubChem exchange data object.

        Args:
            chemicalIdentifier (namedtuple): ChemicalIdentifier(identifierSource, identifierType, identifier)
            exportPath (str, optional): path to export intermediate fetch results (default: None)
            exportIntermediates (bool, optional): flag controlling export of intermediate results (default: False)
            matchIdOnly (bool, optional): flag to stop after performing record lookup to obtain matching identifiers
            contentTypes (list): PubChem content types to fetch (default: ["view", "classification", "property", "xrefs", \
                                                                           "synonyms", "dgidb", "pathway", \
                                                                           "fdaorangebook", "clinicaltrials", "bioactivity"])


        Return:
             (bool, dict): status, PubChem exchange data object

        """
        exportPath = kwargs.get("exportPath", None)
        exportIntermediates = kwargs.get("exportIntermediates", False)
        matchIdOnly = kwargs.get("matchIdOnly", False)
        contentTypes = kwargs.get("contentTypes", ["view", "classification", "property", "xrefs", "synonyms", "dgidb", "pathway", "fdaorangebook", "clinicaltrials", "bioactivity"])

        retStatus = False
        retD = {}
        assemDL = []
        # -- First identifier the PubChem compound CID for the input chemical identifier.
        if chemicalIdentifier.identifierType in ["smiles", "inchi"]:
            searchType = "fastidentity"
        elif chemicalIdentifier.identifierType in ["cid", "name", "inchikey"]:
            searchType = "lookup"
        else:
            pass
        #
        retStatus, rDL = self.fetch(chemicalIdentifier, searchType=searchType, returnType="record", delaySeconds=self.__delaySeconds)
        if not retStatus:
            return retStatus, assemDL
        #
        # PubChem compound CIDs
        cidList = []
        for rD in rDL:
            cid = rD["cid"] if "cid" in rD else None
            if cid:
                cidList.append(cid)
                retD.setdefault(cid, {}).update({"record": rD})
        #
        logger.info(
            "%s searchType %r idType %r idSource %r finds PubChem compounds (%d) %r",
            chemicalIdentifier.idCode,
            searchType,
            chemicalIdentifier.identifierType,
            chemicalIdentifier.identifierSource,
            len(cidList),
            cidList,
        )
        if not cidList:
            return retStatus, assemDL

        if matchIdOnly:
            for cid in retD:
                assemDL.append({"cid": cid, "data": retD[cid]})
            return retStatus, assemDL
        #
        # --  Build data object for each CID
        #
        for cid in cidList:
            chemId = ChemicalIdentifier(idCode=cid, identifier=cid, identifierType="cid")
            for returnType in contentTypes:
                rawResponsePath, extractedResponsePath = None, None
                if exportPath and exportIntermediates:
                    rawResponsePath = os.path.join(exportPath, "%s-pubchem-%s-raw.json" % (cid, returnType))
                    extractedResponsePath = os.path.join(exportPath, "%s-pubchem-%s-extract.json" % (cid, returnType))
                tStatus, rDL = self.fetch(
                    chemId, returnType=returnType, storeRawResponsePath=rawResponsePath, storeResponsePath=extractedResponsePath, delaySeconds=self.__delaySeconds
                )
                # retStatus = tStatus and retStatus
                if tStatus and rDL:
                    if returnType in ["view", "classification", "property", "xrefs", "synonyms"] and len(rDL[0]) > 1:
                        retD[cid][returnType] = rDL[0]
                    else:
                        retD[cid][returnType] = rDL
        #
        for cid in retD:
            assemDL.append({"cid": cid, "data": retD[cid]})
        #
        if exportPath:
            resultPath = os.path.join(exportPath, "%s-pubchem-extracted-all.json" % cid)
            mU = MarshalUtil()
            mU.doExport(resultPath, assemDL, fmt="json", indent=3)

        return retStatus, assemDL

    def fetch(self, chemicalIdentifier, searchType="lookup", returnType="record", storeResponsePath=None, storeRawResponsePath=None, delaySeconds=0.15):
        """Fetch PubChem content (returnType) for the input chemical identifier using the input searchType.

        Args:
            chemicalIdentifer (ChemicalIdentifier Tuple): chemical identifier tuple
            searchType (string): search type (lookup|fastsimilarity|fastidentity)
            returnType (string): object type to return (default: record)

        Return:
            (bool, list): status, return object list (selected items extracted from each returned record type)

        """
        if self.__verbose:
            logger.info(
                "fetching identifier %r nameSpace %r searchType %r returnType %r delay %r",
                chemicalIdentifier.identifier,
                chemicalIdentifier.identifierType,
                searchType,
                returnType,
                delaySeconds,
            )
        #
        if delaySeconds:
            time.sleep(delaySeconds)
        #
        ok = False
        retL = None
        if returnType in ["dgidb", "pathway", "fdaorangebook", "clinicaltrials", "bioactivity"] and chemicalIdentifier.identifierType == "cid":
            response, retCode = self.__doSgdRequest(chemicalIdentifier.identifier, returnType=returnType)
        elif returnType == "view":
            response, retCode = self.__doPugViewRequest(chemicalIdentifier.identifier, nameSpace=chemicalIdentifier.identifierType, domain="compound")
        else:
            response, retCode = self.__doPugRequest(
                chemicalIdentifier.identifier, nameSpace=chemicalIdentifier.identifierType, domain="compound", searchType=searchType, returnType=returnType
            )
        if storeRawResponsePath and response:
            mU = MarshalUtil()
            mU.doExport(storeRawResponsePath, response, fmt="json", indent=3)

        ok = retCode in [200] and response and len(response) > 0
        #
        if ok and returnType == "record":
            retL = self.__parsePubChemRecord(response)
        elif ok and returnType == "xrefs":
            retL = self.__parsePubChemXrefs(response)
        elif ok and returnType == "property":
            retL = self.__parsePubChemProperties(response)
        elif ok and returnType == "classification" and chemicalIdentifier.identifierType == "cid":
            retL = self.__parsePubChemClassifications(chemicalIdentifier.identifier, response)
        elif ok and returnType == "synonyms":
            retL = self.__parsePubChemSynonyms(response)
        elif ok and returnType == "view":
            retL = [self.__parsePubChemCompoundView(response)]
        elif ok and returnType in ["dgidb", "pathway", "fdaorangebook", "clinicaltrials", "bioactivity"]:
            retL = self.__parseSdqResponse(returnType, response)
        #
        if storeResponsePath and retL:
            mU = MarshalUtil()
            mU.doExport(storeResponsePath, retL, fmt="json", indent=3)
        #
        return ok, retL

    def __doPugRequest(self, identifier, nameSpace="cid", domain="compound", searchType="lookup", returnType="record"):
        """Wrapper for PubChem PUG API requests

        https://pubchem.ncbi.nlm.nih.gov/rest/pug/<input specification>/<operation specification>/[<output specification>][?<operation_options>]

        Input:
                <input specification> = <domain>/<namespace>/<identifiers>
                <domain> = substance | compound | assay | <other inputs>
                compound domain <namespace> = cid | name | smiles | inchi | sdf | inchikey | formula | <structure search> | <xref> | listkey | <fast search>
        Operations:
            compound domain <operation specification> = record | <compound property> | synonyms | sids | cids | aids | assaysummary | classification | <xrefs> | description | conformers
            <compound property> = property / [comma-separated list of property tags]

            substance domain <operation specification> = record | synonyms | sids | cids | aids | assaysummary | classification | <xrefs> | description
            <xrefs> = xrefs / [comma-separated list of xrefs tags]

        Examples:
            https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/fastidentity/smiles/C1=NC2=C(N1)C(=O)N=C(N2)N/JSON
            https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/fastsimilarity_2d/smiles/C1=NC2=C(N1)C(=O)N=C(N2)N/JSON?Threshold=99
            https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/property/MolecularFormula,MolecularWeight/CSV
            with “cid=1,2,3,4,5” in the POST body, and  “Content-Type: application/x-www-form-urlencoded”

        """
        #
        httpCodesCatch = [404]
        requestType = "POST"
        outputType = "JSON"
        sslCert = "enabled"
        timeOutSeconds = 10
        retCode, ret = None, None
        try:
            baseUrl = self.__urlPrimary
            pD = {}
            ureq = UrlRequestUtil()
            if domain == "compound":
                if nameSpace in ["cid", "name", "inchikey"] and returnType in ["record"] and searchType in ["lookup"] and requestType == "GET":
                    uId = quote(identifier.encode("utf8"))
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, uId, outputType])
                    ret, retCode = ureq.getUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                elif nameSpace in ["cid", "name", "inchikey"] and returnType in ["record"] and searchType in ["lookup"] and requestType == "POST":
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, outputType])
                    pD = {nameSpace: identifier}
                    ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                #
                elif nameSpace in ["cid", "name", "inchikey"] and returnType in ["xrefs"] and searchType in ["lookup"] and requestType == "POST":
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, returnType, "RegistryID,RN", outputType])
                    pD = {nameSpace: identifier}
                    ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                elif nameSpace in ["cid", "name", "inchikey"] and returnType in ["property"] and searchType in ["lookup"] and requestType == "POST":
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, returnType, "MolecularFormula,XLogP,TPSA,Volume3D", outputType])
                    pD = {nameSpace: identifier}
                    ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                elif nameSpace in ["cid", "name", "inchikey"] and returnType in ["synonyms"] and searchType in ["lookup"] and requestType == "POST":
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, returnType, outputType])
                    pD = {nameSpace: identifier}
                    ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                #
                elif nameSpace in ["cid"] and returnType in ["classification"] and searchType in ["lookup"]:
                    # Needs to be specifically targeted on a particular compound ...
                    uId = quote(identifier.encode("utf8"))
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, uId, returnType, outputType])
                    # pD = {"classification_type": "simple"}
                    pD = {}
                    ret, retCode = ureq.getUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                #
                elif nameSpace in ["cid"] and returnType in ["classification"] and searchType in ["lookup"] and requestType == "POST":
                    # Needs to be specifically targeted on a particular compound ...
                    endPoint = "/".join(["rest", "pug", domain, nameSpace, returnType, outputType])
                    # This is a long request return server codes may be observed 500
                    # pD = {nameSpace: identifier, "classification_type": "simple"}
                    pD = {nameSpace: identifier}
                    ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
                #
                elif nameSpace in ["smiles", "inchi"] and returnType == "record" and searchType in ["fastidentity", "fastsimilarity_2d"] and requestType == "POST":
                    endPoint = "/".join(["rest", "pug", domain, searchType, nameSpace, outputType])
                    if searchType == "fastsimilarity_2d":
                        pD = {nameSpace: identifier, "Threshold": 95}
                    elif searchType == "fastidentity":
                        identityType = "same_stereo_isotope"
                        if identityType in [
                            "same_connectivity",
                            "same_tautomer",
                            "same_stereo",
                            "same_isotope",
                            "same_stereo_isotope",
                            "nonconflicting_stereo",
                            "same_isotope_nonconflicting_stereo",
                        ]:
                            pD = {nameSpace: identifier, "identity_type": identityType}
                        else:
                            pD = {nameSpace: identifier}
                    ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
            elif domain in ["substance", "assay"]:
                logger.error("Fetch not implemented for domain %s", domain)
            #

        except Exception as e:
            logger.error("Failing identifier %r returnType %r nameSpace %r with (retCode %r) %s", identifier, returnType, nameSpace, retCode, str(e))

        return ret, retCode

    def __doPugViewRequest(self, identifier, nameSpace="cid", domain="compound"):
        """Wrapper for PubChem PUG_VIEW API requests

        Example:
            https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/1234/JSON

        """
        requestType = "GET"
        outputType = "JSON"
        retCode, ret = None, None
        httpCodesCatch = [404]
        sslCert = "enabled"
        timeOutSeconds = 10
        try:
            baseUrl = self.__urlPrimary
            #
            pD = {}
            ureq = UrlRequestUtil()
            if nameSpace in ["cid"] and requestType == "GET":
                uId = quote(identifier.encode("utf8"))
                endPoint = "/".join(["rest", "pug_view", "data", domain, uId, outputType])
                ret, retCode = ureq.getUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
            elif nameSpace in ["cid"] and requestType == "POST":
                endPoint = "/".join(["rest", "pug_view", "data", domain, outputType])
                pD = {nameSpace: identifier}
                ret, retCode = ureq.postUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
        except Exception as e:
            logger.error("Failing for identifier %r returnType view with (retCode %r) %s", identifier, retCode, str(e))
        #
        return ret, retCode

    def __doSgdRequest(self, identifier, returnType="dgidb"):
        """Wrapper for PubChem SGD API requests

            Example:

        /sdq/sdqagent.cgi?infmt=json&outfmt=csv&query={"download":"*","collection":"dgidb","where":{"ands":[{"cid":"2244"}]},
                          "order":["relevancescore,desc"],"start":1,"limit":10000000,"downloadfilename":"CID_2244_dgidb"}
        /sdq/sdqagent.cgi?infmt=json&outfmt=csv&query={"download":"*","collection":"pathway","where":{"ands":[{"cid":"2244"},{"core":"1"}]},
                          "order":["name,asc"],"start":1,"limit":10000000,"downloadfilename":"CID_2244_pathway"}

        """
        nameSpace = "cid"
        requestType = "GET"
        retCode, ret = None, None
        httpCodesCatch = [404]
        sslCert = "enabled"
        timeOutSeconds = 10
        try:
            baseUrl = self.__urlPrimary
            pD = {}
            ureq = UrlRequestUtil()
            if nameSpace in ["cid"] and requestType == "GET" and returnType in ["dgidb", "pathway", "fdaorangebook", "clinicaltrials", "bioactivity"]:
                # uId = quote(identifier.encode("utf8"))
                endPoint = "/".join(["sdq", "sdqagent.cgi"])
                pD = {"infmt": "json", "outfmt": "json", "query": '{"select":"*","collection":"%s","where":{"ands":{"cid":"%s"}}}' % (returnType, identifier)}
                ret, retCode = ureq.getUnWrapped(baseUrl, endPoint, pD, httpCodesCatch=httpCodesCatch, returnContentType="JSON", sslCert=sslCert, timeOut=timeOutSeconds)
        except Exception as e:
            logger.error("Failing identifier %r return type %r with (retCode %r) %s", identifier, returnType, retCode, str(e))
        #
        logger.debug("identifier %s returnType %r [%r] result not empty %r", identifier, returnType, retCode, ret is not None)
        #  Display the column names returned from the various SDQ collections -
        # logger.debug("[%r] %s %s", retCode, ret["SDQOutputSet"][0]["collection"], ret["SDQOutputSet"][0]["rows"][0].keys())
        return ret, retCode

    def __parseSdqResponse(self, returnType, vD):
        #
        # Preliminary mapping to capture what is currently deemed useful from these collections.
        #
        cMapD = {
            "dgidb": [
                ("pmids", None),
                ("gid", None),
                ("srcid", None),
                ("geneid", "PubChemGeneId"),
                ("genename", "PubChemGeneName"),
                ("geneclaimname", None),
                ("interactionclaimsource", "provSource"),
                ("interactiontypes", "interactionType"),
                ("sid", "PubChemSid"),
                ("cid", "PubChemCid"),
                ("drugname", "drugName"),
                ("drugclaimname", None),
                ("drugclaimprimaryname", None),
                ("drugchemblid", None),
                ("dois", None),
            ],
            "pathway": [
                ("taxid", "ncbiTaxId"),
                ("pathwayid", "PubChemPathwayId"),
                ("srcid", None),
                ("core", None),
                ("cids", "PubChemCids"),
                ("geneids", "PubChemGeneIds"),
                ("ecs", "enzymeClassifications"),
                ("name", "pathwayName"),
                ("protacxns", None),
                ("taxname", "ncbiTaxonomy"),
                ("pwtype", "pathwayType"),
                ("category", "pathwayCategory"),
                ("url", "pathwayLinkoutUrl"),
                ("source", "provSource"),
                ("extid", "pathwayId"),
                ("externalid", None),
            ],
            "fdaorangebook": [
                ("marketingstatus", "marketing"),
                ("gid", None),
                ("srcid", None),
                ("cid", "PubChemCid"),
                ("activeingredient", "activeIngredient"),
                ("tradename", "tradeName"),
                ("dosageformroute", "formRoute"),
                ("strength", "strength"),
                ("applicationnumber", "FDA Application Number"),
                ("uspatent", "USPatent"),
                ("approvaldate", "FDA Approval Date"),
                ("applicant", "applicant"),
                ("url", "FDAApplicationLinkoutUrl"),
            ],
            "clinicaltrials": [
                ("updatedate", "updateDate"),
                ("gid", None),
                ("srcid", None),
                ("ctid", "clinicalTrialId"),
                ("cids", "PubChemCids"),
                ("link", "clinicalTrialsLinkoutUrl"),
                ("title", "title"),
                ("conditions", "targetConditions"),
                ("interventions", "interventions"),
                ("phase", "trialPhase"),
                ("status", "trialStatus"),
                ("diseaseids", None),
            ],
            "bioactivity": [
                ("cid", "PubChemCid"),
                ("sid", "PubChemSid"),
                ("aid", "PubChemAssayId"),
                ("geneid", "PubChemGeneId"),
                ("taxid", "ncbiTaxonomyId"),
                ("baid", "bioAssayId"),
                ("mid", None),
                ("rnai", None),
                ("hasdrc", None),
                ("aidsrcname", "assaySource"),
                ("ecs", None),
                ("repacxn", None),
                ("aidname", "assayName"),
                ("aidtype", "assayType"),
                ("acvalue", "assayValueMicroMolar"),
                ("aidmdate", "modificationDate"),
                ("activity", "activity"),
                ("protacxn", None),
                ("targetname", "targetName"),
                ("targeturl", "targetUrl"),
            ],
        }
        rL = []
        try:
            mTupL = cMapD[returnType] if returnType in cMapD else []
            for rowD in vD["SDQOutputSet"][0]["rows"]:
                rD = {}
                for srcA, destA in mTupL:
                    if destA and srcA in rowD:
                        rD[destA] = rowD[srcA]
                rL.append(rD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return rL

    def __traversePubChemCompoundView(self, vD):
        """
        "Record": {
                "RecordType": "CID",
                "RecordNumber": 123631,
                "RecordTitle": "Gefitinib",
                "Section": []

        """
        try:
            qD = vD["Record"] if "Record" in vD else {}
            #
            logger.info("keys %s", qD.keys())
            logger.info("Record.RecordType %r", self.__getKeyValues(vD, ["Record.RecordType"]))
            logger.info("Record.RecordNumber %r", self.__getKeyValues(vD, ["Record.RecordNumber"]))
            logger.info("Record.RecordTitle %r", self.__getKeyValues(vD, ["Record.RecordTitle"]))

            for tD in qD["Section"]:
                logger.info("TOCHeading %r Description %r ...", tD["TOCHeading"], tD["Description"][:30])
                if "Section" in tD:
                    for ttD in tD["Section"]:
                        # logger.debug("ttD.keys() %r", ttD.keys())
                        if "Description" in ttD:
                            logger.info("  >> TOCHeading %r Description %r ...", ttD["TOCHeading"], ttD["Description"][:30])
                        else:
                            logger.info("  >> TOCHeading %r Description None ...", ttD["TOCHeading"])
                        if "Section" in ttD:
                            for tttD in ttD["Section"]:
                                if "Description" in tttD:
                                    logger.info("        >> TOCHeading %r Description %r ...", tttD["TOCHeading"], tttD["Description"][:30])
                                else:
                                    logger.info("        >> TOCHeading %r Description  None ...", tttD["TOCHeading"])
                                if "Section" in tttD:
                                    for ttttD in tttD["Section"]:
                                        if "Description" in ttttD:
                                            logger.info("           >> TOCHeading %r Description %r ...", ttttD["TOCHeading"], ttttD["Description"][:30])
                                        else:
                                            logger.info("           >> TOCHeading %r Description  None ...", ttttD["TOCHeading"])
                                        if "Section" in ttttD:
                                            for tttttD in ttttD["Section"]:
                                                if "Description" in tttttD:
                                                    logger.info("              >> TOCHeading %r Description %r ...", tttttD["TOCHeading"], tttttD["Description"][:30])
                                                else:
                                                    logger.info("              >> TOCHeading %r Description  None ...", tttttD["TOCHeading"])

            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return False

    def __parsePubChemCompoundView(self, vD):
        """
        "Record": {
                "RecordType": "CID",
                "RecordNumber": 123631,
                "RecordTitle": "Gefitinib",
                "Section": []

             "Information": [
                       {
                          "ReferenceNumber": 55,
                          "Reference": [
                             "Computed by InChI 1.0.5 (PubChem release 2019.06.18)"
                          ],
                          "Value": {
                             "StringWithMarkup": [
                                {
                                   "String": "InChI=1S/C22H24ClFN4O3/c1-29-20-13-19-16...",
            'Refererence': [
                {
                    "ReferenceNumber": 158,
                    "SourceName": "The National Institute for Occupational Safety and Health (NIOSH)",
                    "SourceID": "npgd0010",
                    "Name": "Acetylsalicylic acid",
                    "Description": "The NIOSH Pocket Guide to Chemical Hazards is intended as a chemicals/Read more: https://www.cdc.gov/niosh/npg/",
                    "URL": "https://www.cdc.gov/niosh/npg/npgd0010.html",
                    "LicenseNote": "The information provided using CDC Web site egulations.",
                    "LicenseURL": "https://www.cdc.gov/Other/disclaimer.html",
                    "ANID": 2266310
                },
                ]                                  }
        """
        pcD = {}
        refD = {}
        mainSectionNameD = {
            "Names and Identifiers": {
                "Computed Descriptors": {"ky": "descriptors", "subSections": []},
                "Other Identifiers": {"ky": "identifiers", "subSections": []},
                "Synonyms": {"ky": "synonyms", "subSections": ["MeSH Entry Terms"]},
            },
            "Chemical and Physical Properties": {
                "Computed Properties": {"ky": "properties", "subSections": []},
                "Experimental Properties": {"ky": "properties", "subSections": []},
            },
            "Drug and Medication Information": {"Drug Indication": {"ky": "drugInfo", "subSections": []}, "Drug Classes": {"ky": "drugInfo", "subSections": []}},
            "Pharmacology and Biochemistry": {
                "Pharmacology": {"ky": "pharmaInfo", "subSections": []},
                "MeSH Pharmacological Classification": {"ky": "pharmaInfo", "subSections": []},
                "FDA Pharmacological Classification": {"ky": "pharmaInfo", "subSections": []},
                "ATC Code": {"ky": "pharmaInfo", "subSections": []},
            },
            # note that classifications are suppressed in the view and must be obtained by other means.
            # "Classification": {"Ontologies": {"ky": "ontologies", "subSections": []}},
        }
        #
        try:
            qD = vD["Record"] if "Record" in vD else {}
            logger.debug("keys %s", qD.keys())
            logger.debug("Record.RecordType %r", self.__getKeyValues(vD, ["Record.RecordType"]))
            logger.debug("Record.RecordNumber %r", self.__getKeyValues(vD, ["Record.RecordNumber"]))
            logger.debug("Record.RecordTitle %r", self.__getKeyValues(vD, ["Record.RecordTitle"]))
            #
            for ref in qD["Reference"]:
                sourceName = ref["SourceName"] if "SourceName" in ref else None
                sourceId = ref["SourceId"] if "SourceId" in ref else None
                name = ref["Name"] if "Name" in ref else None
                description = ref["Description"] if "Description" in ref else None
                url = ref["URL"] if "URL" in ref else None
                licenseInfo = ref["LicenseNote"] if "LicsenceNote" in ref else None
                licenseUrl = ref["LicenseURL"] if "LicenseURL" in ref else None
                refD[ref["ReferenceNumber"]] = {
                    "sourceName": sourceName,
                    "sourceId": sourceId,
                    "name": name,
                    "description": description,
                    "url": url,
                    "licenseInfo": licenseInfo,
                    "licenseUrl": licenseUrl,
                }
            #
            for sectionD in qD["Section"]:
                # Only process main sections identified in mainSectionNameD
                if "TOCHeading" in sectionD and sectionD["TOCHeading"] in mainSectionNameD:
                    sectionName = sectionD["TOCHeading"]
                    sectionMapD = mainSectionNameD[sectionName]
                    logger.debug("Section %r Description %r ...", sectionName, sectionD["Description"][:30])
                    #
                    if "Section" in sectionD:
                        for subSectionD in sectionD["Section"]:
                            subSectionName = subSectionD["TOCHeading"]
                            logger.debug("  >> subSection %r", subSectionName)

                            pcKy = sectionMapD[subSectionName]["ky"] if subSectionD["TOCHeading"] in sectionMapD else None
                            inclSubSections = sectionMapD[subSectionName]["subSections"] if subSectionName in sectionMapD else []
                            # ---
                            if pcKy and "Section" in subSectionD:
                                for subSubSectionD in subSectionD["Section"]:
                                    ky = subSubSectionD["TOCHeading"]
                                    if inclSubSections and ky not in inclSubSections:
                                        logger.debug("        >> SKIPPED subsubSection %r (%r) %r", ky, pcKy, inclSubSections)
                                        continue
                                    logger.debug("        >> subsubSection %r (%r)", ky, pcKy)
                                    # ----
                                    pcD = self.__parseViewInfoList(pcKy, ky, subSubSectionD["Information"], pcD)
                                    #
                            elif pcKy and "Information" in subSectionD:
                                pcD = self.__parseViewInfoList(pcKy, subSectionD["TOCHeading"], subSectionD["Information"], pcD)

                        # ---
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        rD = {}
        for ky in pcD:
            for ky2, pTupS in pcD[ky].items():
                for pTup in pTupS:
                    pD = pTup._asdict()
                    if "provReferenceId" in pD and pD["provReferenceId"] in refD:
                        pD["provData"] = refD[pD["provReferenceId"]]
                    rD.setdefault(ky, {}).setdefault(ky2, []).append(pD)
        return rD

    def __parseViewInfoList(self, pcKy, ky, infoDL, pcD):
        # ----
        #
        try:
            for infD in infoDL:
                referenceNumber = infD["ReferenceNumber"] if "ReferenceNumber" in infD else None
                reference = ",".join(infD["Reference"]) if "Reference" in infD else None
                #
                logger.debug("reference no %r reference %r", referenceNumber, reference)
                name = infD["Name"] if "Name" in infD else None
                if "Value" in infD:
                    if "StringWithMarkup" in infD["Value"]:
                        for sV in infD["Value"]["StringWithMarkup"]:
                            tupV = ExternalAnnotation(name=name, value=sV["String"], dataType="string", featureType=ky, provReferenceId=referenceNumber, provSourceName=reference)
                            pcD.setdefault(pcKy, {}).setdefault(ky, set()).update([tupV])
                    elif "Number" in infD["Value"]:
                        units = infD["Value"]["Unit"] if "Unit" in infD["Value"] else ""
                        for nV in infD["Value"]["Number"]:
                            tupV = ExternalAnnotation(
                                name=name, value=nV, dataType="number", units=units, featureType=ky, provReferenceId=referenceNumber, provSourceName=reference
                            )
                            # tV = str(str(sV) + " " + units).strip()
                            pcD.setdefault(pcKy, {}).setdefault(ky, set()).update([tupV])
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return pcD

    # ---
    def __parsePubChemRecord(self, recordD):
        """Parse selected content from a PubChem PUG API full RECORD response data.

        Args:
            recordD (dict): response from a PubChem PUG API record request with potentially multiple
                            compound records

        Returns:
            list : [{'cid': <pc_cid>, ....}, ...]

        """
        retL = []
        try:
            tDL = recordD["PC_Compounds"] if "PC_Compounds" in recordD else []
            for tD in tDL:
                rD = {}
                tId = self.__getKeyValue(tD, "id.id.cid")
                rD["cid"] = str(tId) if tId else None
                logger.debug("cId = %r", rD["cid"])
                elementCountD = defaultdict(int)
                elements = self.__getKeyValue(tD, "atoms.element")
                for el in elements:
                    elementCountD[str(el)] += 1
                rD["elementCounts"] = elementCountD
                rD["formalCharge"] = self.__getKeyValue(tD, "charge")
                pDL = self.__getKeyValue(tD, "props")
                for pD in pDL:
                    pName = self.__getKeyValue(pD, "urn.name")
                    pLabel = self.__getKeyValue(pD, "urn.label")
                    if pName == "XLogP3" and pLabel == "Log P":
                        rD["XlogP"] = self.__getKeyValue(pD, "value.fval")
                    elif pLabel == "Topological" and pName == "Polar Surface Area":
                        rD["polarSurfaceArea"] = self.__getKeyValue(pD, "value.fval")
                    elif pLabel == "IUPAC Name" and pName == "Preferred":
                        rD["iupacName"] = self.__getKeyValue(pD, "value.sval")
                    elif pLabel == "Molecular Formula":
                        rD["formula"] = self.__getKeyValue(pD, "value.sval")
                    elif pLabel == "InChIKey":
                        rD["InChiKey"] = self.__getKeyValue(pD, "value.sval")

                retL.append(rD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return retL

    def __parsePubChemXrefs(self, recordD):
        """Parse selected content from a PubChem PUG API xrefs response data.

        Args:
            recordD (dict): response from a PubChem PUG API record request with potentially multiple
                            records

        Returns:
            list : [{'cid': <pc_cid>, ....}, ...]

            {
             "InformationList": {
                "Information": [
                    {
                        "CID": 2244,
                        "RegistryID": [
                             ],
                        "RN": [
                        "11126-35-5",
                        ]
                    }
                }
            }
        """

        retL = []
        try:
            tDL = self.__getKeyValue(recordD, "InformationList.Information")
            for tD in tDL:
                rD = {}
                tId = self.__getKeyValue(tD, "CID")
                rD["cid"] = str(tId) if tId else None
                logger.debug("cId = %r", rD["cid"])
                #
                rnL = self.__getKeyValue(tD, "RN")
                rnD = {rn: True for rn in rnL} if rnL else {}
                rIdL = self.__getKeyValue(tD, "RegistryID")
                # This is a laundry list of IDs and it only possible to pull out  clearly prefixed idcodes.
                for rId in rIdL:
                    if rId in rnD:
                        rD.setdefault("CAS", []).append(rId)
                    if rId.startswith("CHEMBL"):
                        rD.setdefault("ChEMBL", []).append(rId)
                    if rId.startswith("CHEBI"):
                        rD.setdefault("ChEBI", []).append(rId)
                    if rId.startswith("HMDB"):
                        rD.setdefault("HMDB", []).append(rId)
                retL.append(rD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return retL

    def __parsePubChemProperties(self, recordD):
        """Parse selected content from a PubChem PUG API properties response data.

        Args:
            recordD (dict): response from a PubChem PUG API record request with potentially multiple
                            records

        Returns:
            list : [{'cid': <pc_cid>, ....}, ...]

                {
                "PropertyTable": {
                    "Properties": [
                        {
                            "CID": 2244,
                            "MolecularFormula": "C9H8O4",
                            "XLogP": 1.2,
                            "TPSA": 63.6,
                            "Volume3D": 136
                        }
                    ]
                }
                }
        """
        retL = []
        try:
            tDL = self.__getKeyValue(recordD, "PropertyTable.Properties")
            for tD in tDL:
                rD = {}
                tId = self.__getKeyValue(tD, "CID")
                rD["cid"] = str(tId) if tId else None
                logger.debug("cId = %r", rD["cid"])
                #
                rD["formula"] = self.__getKeyValue(tD, "MolecularFormula")
                rD["XLogP"] = self.__getKeyValue(tD, "XLogP")
                rD["polarSurfaceArea"] = self.__getKeyValue(tD, "TPSA")
                rD["volume3D"] = self.__getKeyValue(tD, "Volume3D")
                retL.append(rD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return retL

    def __parsePubChemSynonyms(self, recordD):
        """Parse selected content from a PubChem PUG API synonyms response data.

        Args:
            recordD (dict): response from a PubChem PUG API record request with potentially multiple
                            records

        Returns:
            list : [{'cid': <pc_cid>, ....}, ...]
        {
        "InformationList": {
            "Information": [
                {
                    "CID": 2244,
                    "Synonym": [
                    "aspirin",
                    "ACETYLSALICYLIC ACID",
                    "50-78-2",
                    "2-Acetoxybenzoic acid",
                    ]
                },

        """
        retL = []
        try:
            tDL = self.__getKeyValue(recordD, "InformationList.Information")
            for tD in tDL:
                rD = {}
                tId = self.__getKeyValue(tD, "CID")
                rD["cid"] = str(tId) if tId else None
                logger.debug("cId = %r", rD["cid"])
                rD["synonyms"] = self.__getKeyValue(tD, "Synonym")
                retL.append(rD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return retL

    def __parsePubChemClassifications(self, cId, recordD):
        """Parse selected content from a PubChem PUG API classification response data.

        Args:
            recordD (dict): response from a PubChem PUG API record request with potentially multiple
                            records

        Returns:
            list : [{'cid': <pc_cid>, ....}, ...]

            {
            "Hierarchies": {
                "Hierarchy": [
                    {
                    "SourceName": "WHO ATC",
                    "SourceID": "ATCTree",
                    "RootID": "root",
                    "HID": 79,
                    "Information": {
                    "Name": "ATC Code",
                    "Description": [
                        "In the World Health Organization (WHO) Anatomical Therapeutic Chemical (ATC) ..."
                    ],
                    "Comments": [
                        "Drugs are classified in groups at five different levels. The drugs are divided into fourteen main groups .."
                        "Medicinal products are classified according to the main therapeutic use of the main active ingredient, on the  ..."
                    ],
                    "URL": "https://www.whocc.no/atc_ddd_index/",
                    "HNID": 1949731,
                    "ChildID": [
                        "node_1",
                        "node_803",
                        "node_1105",
                        "node_1888",
                        "node_2369",
                        "node_2727",
                        "node_2844",
                        "node_3500",
                        "node_3931",
                        "node_4209",
                        "node_4938",
                        "node_5112",
                        "node_5565",
                        "node_5911"
                    ],
                    "HasCountsOfType": [
                        "CID",
                        "SID"
                    ],
                    "Counts": [
                        {
                            "Type": "CID",
                            "Count": 15837
                        },
                        {
                            "Type": "SID",
                            "Count": 178510
                        }
                    ]
                    },
                    "Node": [
                    {
                        "NodeID": "node_41",
                        "ParentID": [
                            "node_38"
                        ],
                        "Information": {
                            "Name": "A01AD05 - Acetylsalicylic acid",
                            "URL": "https://www.whocc.no/atc_ddd_index/?code=A01AD05",
                            "HNID": 1949772,
                            "Match": true,
                            "Counts": [
                                {
                                "Type": "CID",
                                "Count": 4
                                },
                                {
                                "Type": "SID",
                                "Count": 84
                                }
                            ]
                        }
                    },
                    {
                    "NodeID": "node_38",
                    "ParentID": [
                        "node_3"
                    ],
                    "Information": {
                        "Name": "A01AD - Other agents for local oral treatment",
                        "URL": "https://www.whocc.no/atc_ddd_index/?code=A01AD",
                        "HNID": 1949769,
                        "ChildID": [
                            "node_39",
                            "node_40",
                            "node_41",
                            "node_42",
                            "node_43",
                            "node_44",
                            "node_45"
                        ],
                        "Counts": [
                            {
                            "Type": "CID",
                            "Count": 20
                            },
                            {
                            "Type": "SID",
                            "Count": 303
                            }
                        ]
                    }
                },

        """
        retL = []
        rD = {}
        rD["cid"] = cId
        try:
            tDL = self.__getKeyValue(recordD, "Hierarchies.Hierarchy")
            for tD in tDL:
                sName = self.__getKeyValue(tD, "SourceName")
                sId = self.__getKeyValue(tD, "SourceID")
                logger.debug("Finding sourceName %rsourceId %r", sName, sId)
                #
                provName = self.__getKeyValue(tD, "Information.Name")
                provDescr = self.__getKeyValue(tD, "Information.Description")
                provComments = self.__getKeyValue(tD, "Information.Description")
                provUrl = self.__getKeyValue(tD, "Information.URL")
                provD = {"source": provName, "description": provDescr, "details": provComments, "url": provUrl}
                #
                logger.debug("++++>KEY (%r,%r)", sName, provName)
                if (sName, provName) in [
                    ("WHO ATC", "ATC Code"),
                    ("FDA Pharm Classes", "FDA Pharmacological Classification"),
                    ("IUPHAR/BPS Guide to PHARMACOLOGY", "Target Classification"),
                    # ("ChEMBL", "Target Tree"),
                    # ("ChemIDplus", "ChemIDplus Chemical Information Classification"),
                    ("ChEBI", "ChEBI Ontology"),
                    ("MeSH", "MeSH Tree"),
                ]:
                    nDL = self.__getKeyValue(tD, "Node")
                    cD = {}
                    for nD in nDL:
                        nodeId = self.__getKeyValue(nD, "NodeID")
                        parentL = self.__getKeyValue(nD, "ParentID")
                        childL = self.__getKeyValue(nD, "Information.ChildID")
                        childCount = len(childL) if childL else 0
                        nm = self.__getKeyValue(nD, "Information.Name")
                        descr = self.__getKeyValue(nD, "Information.Description")
                        url = self.__getKeyValue(nD, "Information.URL")
                        idCode = None
                        # --- Add a bit to separate unique identifiers -
                        try:
                            if provName in ["ATC Code"] and nm:
                                fL = nm.split("-")
                                idCode = fL[0].strip()
                                nm = fL[1].strip()
                            elif provName in ["MeSH Tree"] and url:
                                ff = url.split("/")
                                idCode = ff[-1].strip()
                            elif sName == "IUPHAR/BPS Guide to PHARMACOLOGY" and url:
                                ff = url.split("=")
                                idCode = ff[-1].strip()
                        except Exception as e:
                            logger.exception("%s failing with %s", provName, str(e))
                        # ---
                        cD[nodeId] = {"parentList": parentL, "childCount": childCount, "idCode": idCode, "name": nm, "description": descr, "url": url}
                    rD[sName] = {"data": cD, "provenance": provD}
                #
            retL.append(rD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        # --- Process the extracted classification data ----
        try:
            ontClassL = []
            for ret in retL:
                logger.debug("Classification keys %r", ret.keys())
                ontClassD = {}
                ontClassD["cid"] = cId
                for ont, oD in ret.items():
                    childD = {}
                    parentD = {}
                    #
                    if ont in ["cid"]:
                        continue
                    logger.debug("%s node count %d", ont, len(oD["data"]))
                    rootNodeL = []
                    for nm, nD in oD["data"].items():
                        if "parentList" in nD and len(nD["parentList"]) > 1:
                            logger.info("Multiple parents for %s nd %s", ont, nm)
                            continue
                        pN = nD["parentList"][0]
                        if pN not in oD["data"]:
                            rootNodeL.append(nm)
                        childD.setdefault(nm, []).append(pN)
                        parentD.setdefault(pN, []).append(nm)
                    #
                    logger.debug("%s root nodes (%d)", ont, len(rootNodeL))
                    #
                    for ch, prL in childD.items():
                        if len(prL) > 1:
                            logger.info("%s multiple parents for node %s %d", ont, ch, len(prL))
                    # -- enumerate leaf nodes - nodes without children
                    leafL = []
                    for nm in oD["data"]:
                        if nm not in parentD:
                            logger.debug("%s leaf node %r - %s", ont, nm, oD["data"][nm]["name"])
                            leafL.append(nm)
                    #
                    linD = {}
                    for nm in leafL:
                        pN = childD[nm][0] if nm in childD else None
                        linD.setdefault(nm, []).append(nm)
                        while pN and pN in oD["data"]:
                            linD[nm].append(pN)
                            pN = childD[pN][0] if pN in childD else None
                    #
                    for leafN, linL in linD.items():
                        logger.debug("%s: %r : %r %r", ont, leafN, linL, [oD["data"][nm]["name"] for nm in linL])
                    #
                    # Get the unique lineages -
                    #
                    uTupD = {}
                    for leafN, linL in linD.items():
                        linL.reverse()
                        logger.debug("%s linL (%d) %r", ont, len(linL), linL)
                        tupL = [
                            (
                                oD["data"][nm]["name"],
                                oD["data"][nm]["description"][0] if oD["data"][nm]["description"] else None,
                                oD["data"][nm]["idCode"] if oD["data"][nm]["idCode"] else None,
                            )
                            for nm in linL
                        ]
                        logger.debug("%s tupL %r", ont, tupL)
                        tt = tuple(tupL)
                        uTupD[tt] = len(tt)
                    #
                    # Expand this out to
                    for uTup, tLen in uTupD.items():
                        logger.debug("%s (%d) %r", ont, tLen, [u[0] for u in uTup])
                        linL = [{"name": u[0], "description": u[1], "idCode": u[2]} for u in uTup]
                        ontClassD.setdefault(ont, []).append(linL)
                    #
                ontClassL.append(ontClassD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ontClassL

    def __getKeyValues(self, dct, keyNames):
        """Return the tuple of values of corresponding to the input dictionary key names expressed in dot notation.

        Args:
            dct (dict): source dictionary object (nested)
            keyNames (list): list of dictionary keys in dot notatoin

        Returns:
            tuple: tuple of values corresponding to the input key names

        """
        rL = []
        try:
            for keyName in keyNames:
                rL.append(self.__getKeyValue(dct, keyName))
        except Exception as e:
            logger.exception("Failing for key names %r with %s", keyNames, str(e))

        return tuple(rL)

    def __getKeyValue(self, dct, keyName):
        """Return the value of the corresponding key expressed in dot notation in the input dictionary object (nested)."""
        try:
            kys = keyName.split(".")
            for key in kys:
                try:
                    dct = dct[key]
                except KeyError:
                    return None
            return dct
        except Exception as e:
            logger.exception("Failing for key %r with %s", keyName, str(e))

        return None



##
# File:    ResidReader.py
# Author:  J. Westbrook
# Date:    17-Mar-2020
# Version: 0.001
#
# Update:
#
##
"""
Reader for the RESID repository data (RESIDUES.xml).
"""
import logging

logger = logging.getLogger(__name__)


class ResidReader(object):
    """ Reader for the RESID repository data (RESIDUES.xml).
    """

    def __init__(self):
        self.__ns = ""

    def read(self, xTree):
        rL = []
        xRoot = xTree.getroot()
        # <Database id="RESID" release="76.00" date="31-May-2018">
        version = xRoot.attrib.get("release", None)
        for el in xRoot:
            if el.tag == "Entry":
                rL.append(self.__processEntryElement(el))
        return version, rL

    def __processEntryElement(self, entryElement):
        """
        """
        assert entryElement.tag == "{ns}Entry".format(ns=self.__ns)
        #
        doc = {
            "residCode": entryElement.findtext("{ns}Header/{ns}Code".format(ns=self.__ns)),
            "names": [name.text for name in entryElement.findall("{ns}Names/{ns}Name".format(ns=self.__ns))],
            "nameXrefs": [xref.text for xref in entryElement.findall("{ns}Names/{ns}Xref".format(ns=self.__ns))],
            "ontRefs": [xref.text for xref in entryElement.findall("{ns}SequenceCode/{ns}Xref".format(ns=self.__ns))],
            "genEnzymes": [xref.text for xref in entryElement.findall("{ns}GeneratingEnzyme/{ns}EnzymeName".format(ns=self.__ns))],
            "features": [feature.text for feature in entryElement.findall("{ns}Features/{ns}Feature".format(ns=self.__ns))],
        }
        return doc
